from toga.command import GROUP_BREAK, SECTION_BREAK

from .container import TogaContainer
from .libs import Gdk, Gtk


def gtk_toolbar_item_clicked(cmd):
    """Convert a GTK toolbar item click into a command invocation."""

    def _handler(widget):
        cmd.action()

    return _handler


class Window:
    def __init__(self, interface, title, position, size):
        self.interface = interface
        self.interface._impl = self

        self._is_closing = False

        self.layout = None

        self.create()
        self.native._impl = self

        self.native.connect("delete-event", self.gtk_delete_event)

        self.native.set_default_size(size[0], size[1])

        self.set_title(title)
        self.set_position(position)

        # Set the window deletable/closable.
        self.native.set_deletable(self.interface.closable)

        # Added to set Window Resizable - removes Window Maximize button from
        # Window Decorator when resizable == False
        self.native.set_resizable(self.interface.resizable)

        # The GTK window's content is the layout; any user content is placed
        # into the container, which is the bottom widget in the layout. The
        # toolbar (if required) will be added at the top of the layout.
        self.layout = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.native_toolbar = Gtk.Toolbar()
        self.native_toolbar.set_style(Gtk.ToolbarStyle.BOTH)
        self.native_toolbar.set_visible(False)
        self.toolbar_items = {}
        self.layout.pack_start(self.native_toolbar, expand=False, fill=False, padding=0)

        # Because expand and fill are True, the container will fill the available
        # space, and will get a size_allocate callback if the window is resized.
        self.container = TogaContainer()
        self.layout.pack_end(self.container, expand=True, fill=True, padding=0)

        self.native.add(self.layout)

    def create(self):
        self.native = Gtk.Window()

    def get_title(self):
        return self.native.get_title()

    def set_title(self, title):
        self.native.set_title(title)

    def set_app(self, app):
        app.native.add_window(self.native)

    def create_toolbar(self):
        # Remove any pre-existing toolbar content
        if self.toolbar_items:
            self.native_toolbar.set_visible(False)

        for cmd, item_impl in self.toolbar_items.items():
            self.native_toolbar.remove(item_impl)
            try:
                cmd._impl.native.remove(item_impl)
            except AttributeError:
                # Breaks don't have _impls, so there's no native to clean up
                pass

        # Create the new toolbar items
        self.toolbar_items = {}
        for cmd in self.interface.toolbar:
            if cmd == GROUP_BREAK:
                item_impl = Gtk.SeparatorToolItem()
                item_impl.set_draw(True)
            elif cmd == SECTION_BREAK:
                item_impl = Gtk.SeparatorToolItem()
                item_impl.set_draw(False)
            else:
                item_impl = Gtk.ToolButton()
                if cmd.icon:
                    item_impl.set_icon_widget(
                        Gtk.Image.new_from_pixbuf(cmd.icon._impl.native_32)
                    )
                item_impl.set_label(cmd.text)
                if cmd.tooltip:
                    item_impl.set_tooltip_text(cmd.tooltip)
                item_impl.connect("clicked", gtk_toolbar_item_clicked(cmd))
                cmd._impl.native.append(item_impl)
            self.toolbar_items[cmd] = item_impl
            self.native_toolbar.insert(item_impl, -1)

        if self.toolbar_items:
            self.native_toolbar.set_visible(True)
            self.native_toolbar.show_all()

    def set_content(self, widget):
        # Set the new widget to be the container's content
        self.container.content = widget

    def show(self):
        self.native.show_all()

    def hide(self):
        self.native.hide()

    def get_visible(self):
        return self.native.get_property("visible")

    def gtk_delete_event(self, widget, data):
        if self._is_closing:
            should_close = True
        else:
            should_close = self.interface.on_close()

        # Return value of the GTK on_close handler indicates
        # whether the event has been fully handled. Returning
        # False indicates the event handling is *not* complete,
        # so further event processing (including actually
        # closing the window) should be performed.
        return not should_close

    def close(self):
        self._is_closing = True
        self.native.close()

    def get_position(self):
        pos = self.native.get_position()
        return pos.root_x, pos.root_y

    def set_position(self, position):
        self.native.move(position[0], position[1])

    def get_size(self):
        size = self.native.get_size()
        return size.width, size.height

    def set_size(self, size):
        self.native.resize(size[0], size[1])

    def set_full_screen(self, is_full_screen):
        if is_full_screen:
            self.native.fullscreen()
        else:
            self.native.unfullscreen()

    def get_image_data(self):
        display = self.native.get_display()
        display.flush()

        # For some reason, converting the *window* to a pixbuf fails. But if you extract
        # a *part* of the overall screen, that works. So - work out the origin of the
        # window, then the allocation for the container relative to that window, and
        # capture that rectangle.
        window = self.native.get_window()
        origin = window.get_origin()
        allocation = self.container.get_allocation()

        screen = display.get_default_screen()
        root_window = screen.get_root_window()
        screenshot = Gdk.pixbuf_get_from_window(
            root_window,
            origin.x + allocation.x,
            origin.y + allocation.y,
            allocation.width,
            allocation.height,
        )

        success, buffer = screenshot.save_to_bufferv("png")
        if success:
            return buffer
        else:  # pragma: nocover
            # This shouldn't ever happen, and it's difficult to manufacture in test conditions
            raise ValueError(f"Unable to generate screenshot of {self}")
